#include<mpi.h>
#include<stdlib.h>
#include<stdio.h>

#define Wtime MPI_Wtime
#define MEGABYTE (1024*1024)
#define BUFFER_SIZE 1024
#define TAG1 17
#define MASTER_RANK 0
#define MAX_REQUESTS 16
#define MILLION 1000000.0

static int NTIMES = 1000;
static int NTIMES_PREC = 100000;

#define NTIMES_TIMING NTIMES_PREC
#define NTIMES_BARRIER NTIMES
#define NTIMES_LATENCY NTIMES_PREC

#define TEST_SENDRECV 0

#define MASTER if(i_am_the_master) 

int nproc,myid;
int i_am_the_master = 0;

int buf[BUFFER_SIZE],obuf[BUFFER_SIZE];

void Test_MPI_Routines(MPI_Comm comm);
void MPX_Sig_send(int dest,int tag,MPI_Comm comm);
void MPX_Sig_wait(int from,int tag,MPI_Comm comm);

int main(int argc,char *argv[]) {

	int i;
	double t_start,t_end;
	MPI_Init(&argc,&argv);
	MPI_Comm_size(MPI_COMM_WORLD,&nproc);
	MPI_Comm_rank(MPI_COMM_WORLD,&myid);
    i_am_the_master = (myid == MASTER_RANK);

	t_start = Wtime();
	MASTER printf("MPItest/C 1.0: running %d processes...\n",nproc);
	
	if(nproc < 4) {
		MASTER printf("MPItest expects at least 4 processes\n");
		MPI_Finalize();
		exit(0);
	}

	MASTER /* read command line parameters */	
	for(i = 1; i < argc; i ++)
		switch(argv[i][0]) {
			case 'T': NTIMES = atoi(argv[i]+1); break;			
			case 't': NTIMES_PREC = atoi(argv[i]+1); break;			
			default: fprintf(stderr,"WARNING: unrecognized option: %s\n",argv[i]);
			break;
		}

	MPI_Bcast(&NTIMES,1,MPI_INT,MASTER_RANK,MPI_COMM_WORLD);
	MPI_Bcast(&NTIMES_PREC,1,MPI_INT,MASTER_RANK,MPI_COMM_WORLD);

    MASTER printf("Testing basic MPI routines.\n");
	Test_MPI_Routines(MPI_COMM_WORLD);
	t_end = Wtime();
	MASTER printf("MPItest/C complete in %g sec.\n",t_end-t_start);
	MPI_Finalize();
	return 0;
}	

void Test_MPI_Routines(MPI_Comm comm) {
	int nproc,myid,i;
	int i_am_the_master = 0;
	double t, t1, t2, tt, dt, lat, tb;
	MPI_Status status, statuses[MAX_REQUESTS];
	MPI_Request request, requests[MAX_REQUESTS];

	MPI_Comm_size(MPI_COMM_WORLD,&nproc);
	MPI_Comm_rank(MPI_COMM_WORLD,&myid);

    i_am_the_master = (myid == MASTER_RANK);

	MPI_Barrier(comm);

	/* Initialization of buffer data */

	for(i = 0; i < BUFFER_SIZE; i ++)
         buf[i] = myid;

/* ----- Measuring time for a MPI_Wtime() call  ------ */

	MASTER printf("** Timing for MPI operations in microseconds ***\n");

    dt = 0;
	MASTER for(i = 0; i < NTIMES_TIMING; i ++) 
	{
         t1 = Wtime();
         t2 = Wtime();
         dt += (t2 - t1);
	}

	dt /= (double) NTIMES_TIMING;
	t = dt;
	t *= MILLION;

	MASTER printf("%g\t\tTiming\n",t);
	fflush(stdout);
/* ----------- Testing barrier synchronization ---------- */

	tb = 0;
    for(i = 0; i < NTIMES_BARRIER; i ++) 
	{
		 MPI_Barrier(comm);
         t1 = Wtime();
		 MPI_Barrier(comm);
         t2 = Wtime();
         tb += (t2 - t1 - dt);
	}
    tb /= (double) NTIMES_BARRIER;
	t = tb;
	t *= MILLION;

	MASTER printf("%g\t\tBarrier\n",t);
	fflush(stdout);
	
/* ----------- Testing reduction via MPI_Allreduce ---------- */

	t = 0;
    for(i = 0; i < NTIMES; i ++) 
	{
		MPI_Barrier(comm);
		t1 = Wtime();
		MPI_Allreduce(buf,obuf,1,MPI_INT,MPI_SUM,comm);
		MPI_Barrier(comm);
		t2 = Wtime();
		t += (t2 - t1 - dt - tb);
	}
	t /= (double) NTIMES;
	t *= MILLION;

	MASTER printf("%g\t\tGlobal sum (Allreduce)\n",t);
	fflush(stdout);
/* ----------- Testing vector reduction via MPI_Allreduce ---------- */

	t = 0;
    for(i = 0; i < NTIMES; i ++) 
	{
		MPI_Barrier(comm);
		t1 = Wtime();
		MPI_Allreduce(buf,obuf,10,MPI_INT,MPI_SUM,comm);
		MPI_Barrier(comm);
		t2 = Wtime();
		t += (t2 - t1 - dt - tb);
	}
	t /= (double) NTIMES;
	t *= MILLION;

	MASTER printf("%g\t\t10 global sums (Allreduce)\n",t);
	fflush(stdout);
/* ----------- Testing reduction via MPI_Reduce ---------- */
	t = 0;
    for(i = 0; i < NTIMES; i ++) 
	{
		MPI_Barrier(comm);
		t1 = Wtime();
		MPI_Reduce(buf,obuf,1,MPI_INT,MPI_SUM,MASTER_RANK,comm);
		MPI_Barrier(comm);
		t2 = Wtime();
		t += (t2 - t1 - dt - tb);
	}
	t /= (double) NTIMES;
	t *= MILLION;

	MASTER printf("%g\t\tGlobal sum (Reduce)\n",t);
	fflush(stdout);
/* ----------- Testing vector reduction via MPI_Reduce ---------- */
	t = 0;
    for(i = 0; i < NTIMES; i ++) 
	{
		MPI_Barrier(comm);
		t1 = Wtime();
		MPI_Reduce(buf,obuf,10,MPI_INT,MPI_SUM,MASTER_RANK,comm);
		MPI_Barrier(comm);
		t2 = Wtime();
		t += (t2 - t1 - dt - tb);
	}
	t /= (double) NTIMES;
	t *= MILLION;

	MASTER printf("%g\t\t10 Global sums (Reduce)\n",t);
	fflush(stdout);
/* ----------- Testing broadcast routine ---------- */

	t = 0;
    for(i = 0; i < NTIMES; i ++) 
	{
		MPI_Barrier(comm);
		t1 = Wtime();
		MPI_Bcast(buf,1,MPI_INT,MASTER_RANK,comm);
		MPI_Barrier(comm);
		t2 = Wtime();
		t += (t2 - t1 - dt - tb);
	}
	t /= (double) NTIMES;
	t *= MILLION;

	MASTER printf("%g\t\tBroadcast\n",t);
	fflush(stdout);
/* ----------- Testing gather routine ---------- */

	t = 0;
    for(i = 0; i < NTIMES; i ++) 
	{
		MPI_Barrier(comm);
		t1 = Wtime();
		MPI_Gather(buf,1,MPI_INT,obuf,1,MPI_INT,MASTER_RANK,comm);
		MPI_Barrier(comm);
		t2 = Wtime();
		t += (t2 - t1 - dt - tb);
	}
	t /= (double) NTIMES;
	t *= MILLION;

	MASTER printf("%g\t\tGather\n",t);
	fflush(stdout);
/* ----------- Testing nonblocking send ---------- */

	t = 0;
	if(i_am_the_master)
	    for(i = 0; i < NTIMES; i ++) 
		{
			t1 = Wtime();
			MPI_Isend(buf,10,MPI_INT,1,TAG1,comm,&request);
			t2 = Wtime();
			t += (t2 - t1 - dt);
		}
	else if(myid == 1)
	    for(i = 0; i < NTIMES; i ++) 
		{
			MPI_Irecv(obuf,10,MPI_INT,MASTER_RANK,TAG1,comm,&request);
			MPI_Wait(&request,&status);
		}
	t /= (double) NTIMES;
	t *= MILLION;

	MASTER printf("%g\t\tNon-blocking vector send\n",t);
	fflush(stdout);
/* ----------- Testing non-blocking vector send with wait ---------- */

	t = 0;
	if(i_am_the_master)
	    for(i = 0; i < NTIMES; i ++) 
		{
			t1 = Wtime();
			MPI_Isend(buf,10,MPI_INT,1,TAG1,comm,&request);
			MPI_Wait(&request,&status);

			t2 = Wtime();
			t += (t2 - t1 - dt);
		}
	else if(myid == 1)
	    for(i = 0; i < NTIMES; i ++) 
		{
			MPI_Irecv(obuf,10,MPI_INT,MASTER_RANK,TAG1,comm,&request);
			MPI_Wait(&request,&status);
		}
	t /= (double) NTIMES;
	t *= MILLION;

	MASTER printf("%g\t\tNon-blocking vector send with wait\n",t);
	fflush(stdout);
/* ----------- Testing blocking vector send ---------- */

	t = 0;
	if(i_am_the_master)
	    for(i = 0; i < NTIMES; i ++) 
		{
			t1 = Wtime();
			MPI_Send(buf,10,MPI_INT,1,TAG1,comm);
			t2 = Wtime();
			t += (t2 - t1 - dt);
		}
	else if(myid == 1)
	    for(i = 0; i < NTIMES; i ++) 
			MPI_Recv(obuf,10,MPI_INT,MASTER_RANK,TAG1,comm,&status);

	t /= (double) NTIMES;
	t *= MILLION;

	MASTER printf("%g\t\tBlocking vector send\n",t);
	fflush(stdout);

/* ----------- Testing simultaneous send & receive ---------- */

	t = 0;
	if(i_am_the_master)
	    for(i = 0; i < NTIMES; i ++) 
		{
			t1 = Wtime();
            MPI_Sendrecv(buf,1,MPI_INT,1,TAG1,
                         obuf,1,MPI_INT,1,TAG1,
                         comm,statuses);
			t2 = Wtime();
			t += (t2 - t1 - dt);
		}
	else if(myid == 1) 
	    for(i = 0; i < NTIMES; i ++) 
            MPI_Sendrecv(buf,1,MPI_INT,MASTER_RANK,TAG1,
                         obuf,1,MPI_INT,MASTER_RANK,TAG1,
                         comm,statuses);

	t /= (double) NTIMES;
	t *= MILLION;

	MASTER printf("%g\t\tSendRecv\n",t);
	fflush(stdout);

/* ----------- Testing sequential send & receive ---------- */
	t = 0;
	if(i_am_the_master)
	    for(i = 0; i < NTIMES; i ++) 
		{
			t1 = Wtime();
            MPI_Send(buf,1,MPI_INT,1,TAG1,comm);
            MPI_Recv(obuf,1,MPI_INT, 1,TAG1,comm,&status);
			t2 = Wtime();
			t += (t2 - t1 - dt);
		}
	else if(myid == 1) 
	    for(i = 0; i < NTIMES; i ++) 
		{
            MPI_Recv(obuf,1,MPI_INT,MASTER_RANK,TAG1,comm,&status);
            MPI_Send(buf,1,MPI_INT,MASTER_RANK,TAG1,comm);
		}
	t /= (double) NTIMES;
	t *= MILLION;

	MASTER printf("%g\t\tSend and receive\n",t);
	fflush(stdout);
/* ----------- Testing simultaneous execution of requests ---------- */

	t = 0;
	if(i_am_the_master)
	    for(i = 0; i < NTIMES; i ++) 
		{
            t1 = Wtime();
            MPI_Send_init(buf,1,MPI_INT,1,TAG1,comm,&requests[0]);            
            MPI_Send_init(buf,1,MPI_INT,1,TAG1,comm,&requests[1]);
            MPI_Send_init(buf,1,MPI_INT,1,TAG1,comm,&requests[2]);
            MPI_Send_init(buf,1,MPI_INT,1,TAG1,comm,&requests[3]);
            MPI_Send_init(buf,1,MPI_INT,1,TAG1,comm,&requests[4]);
            MPI_Startall(5,requests);
            MPI_Waitall(5,requests,statuses);
            t2 = Wtime();
			t += (t2 - t1 - dt);
		}
	else if(myid == 1) 
	    for(i = 0; i < NTIMES; i ++) 
		{
            MPI_Recv_init(&obuf[0],1,MPI_INT,0,TAG1,comm,&requests[0]);
            MPI_Recv_init(&obuf[1],1,MPI_INT,0,TAG1,comm,&requests[1]);
            MPI_Recv_init(&obuf[2],1,MPI_INT,0,TAG1,comm,&requests[2]);
            MPI_Recv_init(&obuf[3],1,MPI_INT,0,TAG1,comm,&requests[3]);
            MPI_Recv_init(&obuf[4],1,MPI_INT,0,TAG1,comm,&requests[4]);
            MPI_Startall(5,requests);
            MPI_Waitall(5,requests,statuses);
		}
	t /= (double) NTIMES;
	t *= MILLION;

	MASTER printf("%g\t\t5 sends in one\n",t);
	fflush(stdout);
/* ----------- Testing execution of requests ---------- */

	t = 0;
	if(i_am_the_master)
	    for(i = 0; i < NTIMES; i ++) 
		{
            t1 = Wtime();
            MPI_Isend(buf,1,MPI_INT,1,TAG1,comm,&requests[0]);            
            MPI_Isend(buf,1,MPI_INT,1,TAG1,comm,&requests[1]);            
            MPI_Isend(buf,1,MPI_INT,1,TAG1,comm,&requests[2]);            
            MPI_Isend(buf,1,MPI_INT,1,TAG1,comm,&requests[3]);            
            MPI_Isend(buf,1,MPI_INT,1,TAG1,comm,&requests[4]);            
            MPI_Waitall(5,requests,statuses);
            t2 = Wtime();
			t += (t2 - t1 - dt);
		}
	else if(myid == 1) 
	    for(i = 0; i < NTIMES; i ++) 
		{
            MPI_Recv_init(&obuf[0],1,MPI_INT,0,TAG1,comm,&requests[0]);
            MPI_Recv_init(&obuf[1],1,MPI_INT,0,TAG1,comm,&requests[1]);
            MPI_Recv_init(&obuf[2],1,MPI_INT,0,TAG1,comm,&requests[2]);
            MPI_Recv_init(&obuf[3],1,MPI_INT,0,TAG1,comm,&requests[3]);
            MPI_Recv_init(&obuf[4],1,MPI_INT,0,TAG1,comm,&requests[4]);
            MPI_Startall(5,requests);
            MPI_Waitall(5,requests,statuses);
		}

	t /= (double) NTIMES;
	t *= MILLION;

 	MASTER printf("%g\t\t5 sends\n",t);
	fflush(stdout);
/* ----------- Testing signal send/receive ---------- */
	t = 0;
	if(i_am_the_master)
	    for(i = 0; i < NTIMES_LATENCY; i ++) 
		{
            t1 = Wtime();
            MPX_Sig_send(1,TAG1,comm);
            MPX_Sig_wait(1,TAG1,comm);
            t2 = Wtime();
			t += (t2 - t1 - dt);
		}
	else if(myid == 1) 
	    for(i = 0; i < NTIMES_LATENCY; i ++) 
		{
            MPX_Sig_wait(MASTER_RANK,TAG1,comm);
            MPX_Sig_send(MASTER_RANK,TAG1,comm);            
		}
		
	t /= (double) NTIMES_LATENCY;
	t *= MILLION;

	MASTER printf("%g\t\tsignal sending (latency)\n",t);
	fflush(stdout);	
}

void MPX_Sig_send(int dest,int tag,MPI_Comm comm) {
	MPI_Send(NULL,0,MPI_INT,dest,tag,comm);
}

void MPX_Sig_wait(int from,int tag,MPI_Comm comm) {
	MPI_Status status;
	MPI_Recv(NULL,0,MPI_INT,from,tag,comm,&status);
}

